<?
/** -----------------------------------------------------------------------------------------------
 * @{
 * Functions library to communicate with the XMPP Server of the Logitech Harmony Hub.
 *
 * V0.7 - 01/04/2016
 * by Zapp for the IP-Symcon community.
 * 
 * Based on protocol descriptions and code from Pete LePage (https://petelepage.com/blog/)
 * and Jeff Terrace. Pyton code available at https://github.com/petele/pyharmony
 *
 * Prerequisites:
 * - A MyHarmony account and a configured Harmony Hub are required.
 * - The Hub must have a fixed IP
 * - IPSInstaller library (IPSInstaller.ips.php). Update: Not needed.
 *
 * Installation & Configuration
 * - Create a Client Socket Instance in IPS on port 5222 of the Hub IP
 * - Create a dedicated Category in IPS (for example "Harmony")
 * - Create an empty script in this category and copy the content of "Harmony_Read.ips.php" into it.
 * - Create a Register Variable. Set the Client Socket Instance as Parent and the "Harmony_Read" script as Target ID. 
 * - Indicate the Hub IP Address and Client Socket Instance ID in the configuration below.
 * - Indicate your MyHarmony login (email) and password in the configuration below. 
 *
 * Steps:
 * 1) Harmony_init()
 *	  	- Retrives the authentication token based on email and password from the Logitech server
 *		  The retrieval of the authentication token is needed only once.
 *   	  It is stored in an automatically created IPS Variable for subsequent connections.
 * 		- Connects to the Hub XMPP Server as guest 
 *		- Note: The subsequent stage to connect as an authenticated user seems not to be required
 * 2) Harmony_getConfig()
 *		- Sends a request to the server to send the configuration.
 *		- The configuration is sent by the server under the form of an <IQ><OA> message.
 *		- The read script ("Harmony_Read.ips.php") stores the config in a file using Harmony_processIQ().
 * 3) Harmony_formatConfig()
 *		- Reads the file with the stored config in it.
 *		- Prints the config out in the IPS output
 * 4) Send commands
 *
 * Example commands:
 * 		Harmony_sendCommand("13831185", "PowerOff"); // TV Samsung - Off
 *		Harmony_startActivity(-1);	// Switch all devices from current activity off ("off" button)
 *
 * Known issues:
 * - The Client Socket to the server regularly disconnects. Commands need to be sent again.
 *
 * ChangeLog v0.7:
 * - Harmony_formatConfig(): Updated method for parsing Activities and Devices from the configuration file.
 * - Added debugging functionalities.
 *
 * ChangeLog v0.6:
 * - Updated for last Harmony protocol for config
 * - Harmony_formatConfig() creates boolean Variables for each Activity. This Variables are set to true if the corresponding Activity is started using Harmony_setActivityVariable().
 * - Harmony_getActivityFromId() returns a String activty description from the Harmony Activity ID.
 *
 * ChangeLog v0.5:
 * - Harmony Hub Socket is connected if in disconnected status in XMPP_Send() - Thanks to Xackery (http://www.ip-symcon.de/forum/threads/22682-Logitech-Harmony-Ultimate-amp-Smart-Control-Hub-library?p=213639#post213639)
 *
 * Changelog v0.4:
 * - Fixed bug in authentication call which prevented identity from being returned
 * - Added resource binding
 * - Implemented proper XMPP client/server interaction
 *
 * Changelog v0.3:
 * - XMPP_getPayload() returns an array with type of message, associated activity and error codes
 * - Harmony_processIQ() automatically starts Harmony_formatConfig() when aggregation of message is finished
 * - Harmony_formatConfig() creates an array with Actvitiy->ID and stores it in a IPS Variable
 *
 * @file          Harmony_XMPP.ips.php
 * @author        Zapp
 *
 */
// ---------------------------------------------------------------------------------------------
$debug = true;

$LOGITECH_AUTH_URL = "https://svcs.myharmony.com/CompositeSecurityServices/Security.svc/json/GetUserAuthToken";
$HARMONY_CLIENT_INSTANCE_ID = 12345 /*[Harmony XMPP CLient]*/;
$HARMONY_CLIENT_RESOURCE = 'ips';  // gatorade. ?

// MyHarmony user credentials
$email = "user@domain.com";
$password = "password";

// File to write the Harmony config
$configFilePath = IPS_GetKernelDir()."Harmony_Config.txt";

// IPS Variable Profile Name for Harmony Activity
$profileNameHarmonyActivity = ".Harmony.Activity";

// This 'from' value can be retrieved from the messages received by the Server ("to" field). But it does not seem necessary.
$from = '';

//*****************************************************
// Do not edit below this point
//*****************************************************

/**
* Opens the stream to XMPP Server
*
* @return none
**/
function XMPP_OpenStream() {
   XMPP_Send("<stream:stream to='connect.logitech.com' xmlns:stream='http://etherx.jabber.org/streams' xmlns='jabber:client' xml:lang='en' version='1.0'>"); //  xmlns:xml="http://www.w3.org/XML/1998/namespace"
}

/**
* Closes the stream to XMPP Server
*
* @return none
**/
function XMPP_CloseStream() {
	XMPP_Send("</stream:stream>");  // <presence type='unavailable'/>
}

/**
* Sends Auth request to XMPP server
*
* @param string $user
* @param string $password
*
* @return none
**/
function XMPP_Auth($user, $password) {
	global $debug;
	
	if ($debug) IPS_LogMessage("HARMONY XMPP", "DEBUG: Authenticating with $user - $password");
	$pass = base64_encode("\x00" . $user . "\x00" . $password);
	XMPP_Send("<auth xmlns='urn:ietf:params:xml:ns:xmpp-sasl' mechanism='PLAIN'>" . $pass . "</auth>");
}

/**
* Sends Bind request to XMPP server
*
* @param string $resource A resource name
*
* @return none
**/
function XMPP_Bind($resource) {
	global $debug;
	if ($debug) IPS_LogMessage("HARMONY XMPP", "DEBUG: Binding with resource $resource");
	XMPP_Send("<iq type='set' id='bind_2'><bind xmlns='urn:ietf:params:xml:ns:xmpp-bind'><resource>$resource</resource></bind></iq>");
}

/**
* Sends Session request to XMPP server
*
* @return none
**/
function XMPP_Session() {
	global $debug;
	if ($debug) IPS_LogMessage("HARMONY XMPP", "DEBUG: Sending Session request");
	XMPP_Send("<iq id='bind_3' type='set'><session xmlns='urn:ietf:params:xml:ns:xmpp-session'/></iq>");
}

/**
* Sends XMPP message to XMPP server
* If the socket is closed, adds the command to a queue array for subsequent execution.
*
* @param string $payload
*
* @return boolean true if success
**/
function XMPP_Send($payload) {
	global $debug, $HARMONY_CLIENT_INSTANCE_ID;
	
	$instanceHarmonySocket = IPS_GetInstance($HARMONY_CLIENT_INSTANCE_ID);
	// Open the socket if it is disconnected
	if ($instanceHarmonySocket['InstanceStatus'] != 102) {
		CSCK_SetOpen($HARMONY_CLIENT_INSTANCE_ID, true);
		IPS_ApplyChanges($HARMONY_CLIENT_INSTANCE_ID);
	}
		$result = CSCK_SendText($HARMONY_CLIENT_INSTANCE_ID , $payload);
		if ($result) {
			if ($debug) {
				echo "SUCCESS message sent\r\n";
				return true;
			} 
		} else {
			IPS_LogMessage("HARMONY XMPP", "ERROR in XMPP_Send(): Sending XMPP message failed (Socket was connected).");
			echo "ERROR sending message\r\n";
			return false;
		}
	// Open the socket after the command is sent if it is disconnected
	if ($instanceHarmonySocket['InstanceStatus'] != 102) {
		CSCK_SetOpen($HARMONY_CLIENT_INSTANCE_ID, true);
		IPS_ApplyChanges($HARMONY_CLIENT_INSTANCE_ID);
	}
	// The Code below is an attempt to implement a command queue in case the socket is not connected - Work in Progress (05/01/2014)
	// REMOVED
}

/**
* Extracts CDATA payload from XMPP xml message
*
* @param string $xml
*
* @return array CDATA content formatted as 'type': Type of message, 'activityId', 'errorCode', 'errorString'
**/
function XMPP_getPayload($xml) {
	global $debug;
	
	preg_match('/type="[a-zA-Z\.]+\?(.*)">/', $xml, $type);  // type="harmony.engine?startActivityFinished"
	
	preg_match('/<!\[(CDATA)\[\s*(.*?)\s*\]\]>/', $xml, $cdata);
	//if (strpos('{', $cdata) == 1) ...;
	if (preg_match('/^{(.*?)}/', $cdata[2], $cdatat) == 1) {
		$content = explode(',', $cdatat[1]);
	} else {
		$content = explode(':', $cdata[2]);
	}
	
	$items['type'] = $type[1];
	foreach ($content as $item) {
		$itemParts = explode('=', $item);
		$items[$itemParts[0]] = $itemParts[1];
	}
	
	return $items;
}

/**
* Initializes connection to Harmony Hub XMPP Server
* This includes getting the User Auth token from the remote Logitech Server if needed and then authenticating to the XMPP server
*
* @return true if authentication succeeded
**/
function Harmony_init() {
	global $debug, $email, $password;

	if ($debug) echo "-> Checking for User Auth Token...\r\n";
	$parentID = IPS_GetParent($_IPS['SELF']);
	$tokenVariableId = @IPS_GetVariableIDByName("User Auth Token", $parentID);
	if ($tokenVariableId === false) {
		$tokenVariableId = IPS_CreateVariable(3);
		IPS_SetName($tokenVariableId, "User Auth Token");
		IPS_SetParent($tokenVariableId, $parentID);
		IPS_SetHidden($tokenVariableId, true);
		$UserAuthToken = Harmony_getAuthToken($email, $password);
		SetValue($tokenVariableId, $UserAuthToken);
	} else {
		$UserAuthToken = GetValue($tokenVariableId);
		if ($UserAuthToken == "") {
			$UserAuthToken = Harmony_getAuthToken($email, $password);
			SetValue($tokenVariableId, $UserAuthToken);
		}
	}
	if ($debug) echo "--> UserAuthToken: $UserAuthToken\r\n";
	
	if ($UserAuthToken != "") {
		if ($debug) echo "--> Requesting stream...\r\n";
		XMPP_OpenStream();
	} else {
		IPS_LogMessage("HARMONY XMPP", "ERROR in Harmony_init(): No Auth Token!");
		echo "--> ERROR: No Auth Token!\r\n";
		return false;
	}

	
}

/**
* Gets User Auth Token from remote Logitech Auth Server
*
* @param string $email
* @param string $password
*
* @return string Auth Token from Logitech server
**/
function Harmony_getAuthToken($email, $password) {
	global $debug, $LOGITECH_AUTH_URL;
	$timeout = 30;

	if ($debug) echo "-> Retrieving Auth Token from My Harmony Account Server:\r\n";
	//$headers = {'content-type': 'application/json; charset=utf-8'}
	$credentials = array(
				'email' => $email,
				'password' => $password
				);
	$json_string = json_encode($credentials);// '{'.$cmd.'}';

	$ch = curl_init($LOGITECH_AUTH_URL);

	$options = array(
               CURLOPT_TIMEOUT => $timeout,
               CURLOPT_CONNECTTIMEOUT => $timeout,
               CURLOPT_VERBOSE => 1,
               CURLOPT_SSL_VERIFYPEER => false,
               CURLOPT_SSL_VERIFYHOST => 2,
               CURLOPT_RETURNTRANSFER => true,
               CURLOPT_HTTPHEADER => array('Content-type: application/json; charset=utf-8'),
               CURLOPT_POST => true,
               CURLOPT_POSTFIELDS => $json_string
               );

	// Setting curl options
	curl_setopt_array( $ch, $options );
	// Getting results
	$json_result =  curl_exec($ch);
	if ($json_result === FALSE) {
    	die(curl_error($ch));
	}
	$result = json_decode($json_result);

	if(!curl_errno($ch)){
  		// No Error
	} else {
  		echo "	ERROR: Curl failed - " . curl_error($ch);
  		IPS_LogMessage("HARMONY XMPP", "ERROR: Authentification failed");
	}

	//print_r ($result);
	$UserAuthToken = $result->GetUserAuthTokenResult->UserAuthToken;

	return $UserAuthToken;
}

/**
* Sends a request to swap Auth Token for a Session token to the XMPP Server
* The returned IQ Message with the Session token is processed by Harmony_processIQ().
*
* @return none
**/
function Harmony_sendSessionTokenRequest($token) {
	$tokenString = $token.":name=foo#iOS6.0.1#iPhone"; // "token=".
	XMPP_Send("<iq type='get' id='3174962747' from='guest'><oa xmlns='connect.logitech.com' mime='vnd.logitech.connect/vnd.logitech.pair'>token=$tokenString</oa></iq>");
}

/**
* Sends request to get Harmony configuration to XMPP Server
* The server will return the xml encoded config in a IQ/OA reply.
*
* @return none
**/
function Harmony_getConfig() {
	global $from;
	
	$iqString = "<iq type='get' id='2320426445' from='$from'>
      <oa xmlns='connect.logitech.com' mime='vnd.logitech.harmony/vnd.logitech.harmony.engine?config'>
      </oa>
    </iq>";
    XMPP_Send($iqString);
}

/**
* Sends a request to the XMPP Server to get the current Activity ID
*
* @return none
**/
function Harmony_getCurrentActivity() {
	global $from;
	
	$iqString = "<iq type='get' id='2320426445' from='$from'>
      <oa xmlns='connect.logitech.com' mime='vnd.logitech.harmony/vnd.logitech.harmony.engine?getCurrentActivity'>
      </oa>
    </iq>";
    XMPP_Send($iqString);
}

/**
* Sends request to send an IR command to XMPP Server
* Device ID and Command name have to be retrieved from the config. No error check is made.
*
* @param $deviceID ID as retrieved from the Harmony config
* @param $command command as retrieved from teh Harmony config
*
* @return none
**/
function Harmony_sendCommand($deviceID, $command) {
   $iqString = "<iq type='get' id='5e518d07-bcc2-4634-ba3d-c20f338d8927-2'><oa xmlns='connect.logitech.com' mime='vnd.logitech.harmony/vnd.logitech.harmony.engine?holdAction'>action={\"type\"::\"IRCommand\",\"deviceId\"::\"$deviceID\",\"command\"::\"$command\"}:status=press</oa></iq>";
    XMPP_Send($iqString);
}

/**
* Sends request to send an IR command to start a given activity to the XMPP Server
* The Activity ID has to be retrieved from the config. No error check is made.
*
* @param $activityID ID as retrieved from the Harmony config
*
* @return none
**/
function Harmony_startActivity($activityID) {
   $iqString = "<iq type='get' id='5e518d07-bcc2-4634-ba3d-c20f338d8927-2'><oa xmlns='connect.logitech.com' mime='vnd.logitech.harmony/vnd.logitech.harmony.engine?startactivity'>activityId=$activityID:timestamp=0</oa></iq>";
   XMPP_Send($iqString);

}

/**
* Sets the Activity Variables status in IPS (Alternative 1).
*
* @param $activityID ID as retrieved from the Harmony config
*
* @return true if Variable exists and has been changed to $status
**/
function Harmony_setActivityVariables($activityID, $status) {
	$activity = Harmony_getActivityFromId($activityID);
	
	$parentID = IPS_GetParent($_IPS['SELF']);
	$CatID = @IPS_GetCategoryIDByName("Activities", $parentID);
	// Reset all Variables to false
	foreach (IPS_GetChildrenIDs($CatID) as $id) {
		SetValue($id, false);
	}
	// Set the Specific Variable to $status
	if ($CatID !== false) {
		$activitiyVariableId = @IPS_GetVariableIDByName($activity." Activity", $CatID);
		if ($activitiyVariableId !== false) {
			SetValue($activitiyVariableId, true);
			return true;
		} else {
			return false;
		}
	} else {
		return false;
	}
}

/**
* Sets the Activity Variable status in IPS (Alternative 2).
*
* @param $activityID ID as retrieved from the Harmony config
*
* @return true if Variable exists and has been changed
**/
function Harmony_setActivityVariable($activityID) {
	$parentID = IPS_GetParent($_IPS['SELF']);
	$activitiyVariableId = @IPS_GetVariableIDByName("Activity", $parentID);
	if ($activitiyVariableId !== false) {
		SetValue($activitiyVariableId, $activityID);
		return true;
	} else {
		return false;
	}
}

/**
* Formats the configuration received from XMPP Server with getConfig and stored in an IPS String Variable
*
* @param $displayOnly true to display only the Activities/Devices list, fals to also write the list into IPS Variables. 
* 		 WARNING: This should be set only for debugging. If the output is not written to the IPS Variables, setActivity() will not work.
*
* @return config array or false if error
**/
function Harmony_formatConfig($displayOnly = false) {
	global $debug, $configFilePath, $profileNameHarmonyActivity;
	
	$parentID = IPS_GetParent($_IPS['SELF']);
	
	$configVariableId = @IPS_GetVariableIDByName("Harmony Config", IPS_GetParent($_IPS['SELF']));
	if ($configVariableId === false) {
		IPS_LogMessage("HARMONY XMPP", "ERROR in Harmony_formatConfig(): Config Variable does not exist!");
		return false;
	} else {
		// The approach with a variable seems not to work?
		$config = file_get_contents($configFilePath);
		
		// Find Activity & Devices names ($tags[1]) and IDs ($tags[2])
		echo "HARMONY CONFIGURATION ";
		if ($displayOnly) echo "(display only)";
		echo ":\r\n";
		preg_match_all('/,\"label\":\"([a-z \-A-Z0-9]+)\",\"id\":\"([0-9]+)\",\"([a-z A-Z]+)\"/', $config, $tags);  // Update 12/03/2016: Updated first capturing group for special signs and added new group
		
		// Now for each ID, find the available commands
		$tagId = 0;
		$activities = array();
		if (!$displayOnly) {
			// Alternative 2: Create a Variable Profile for the Harmony Activities
			if (IPS_VariableProfileExists($profileNameHarmonyActivity)) {
				@IPS_DeleteVariableProfile($profileNameHarmonyActivity);
			}
			@IPS_CreateVariableProfile($profileNameHarmonyActivity, 1);
			IPS_SetVariableProfileText($profileNameHarmonyActivity, "", "");
			IPS_SetVariableProfileValues($profileNameHarmonyActivity, 0, 0, 0);
			IPS_SetVariableProfileDigits($profileNameHarmonyActivity, 0);
			//IPS_SetVariableProfileIcon(".Harmony.Activity", $Icon);
		}
		foreach($tags[2] as $id) {			
				
			if ($tags[3][$tagId] == "Capabilities") {  // Update 01/04/2016: It seems that Activities ID size can now also be > 7 (was = 7 before and Devices ID size = 8)
				echo "\r\nDevice: ".$tags[1][$tagId]." -> Id: ".$tags[2][$tagId]."\r\n";
				if ($displayOnly) echo " debug:  tag3: ".$tags[3][$tagId]."\r\n";  // For debug
				
				echo "  Available Commands for ".$id.":\r\n";
				preg_match_all('/\\\"command\\\":\\\"([a-z A-Z0-9]+)\\\",\\\"type\\\":\\\"IRCommand\\\",\\\"deviceId\\\":\\\"'.$id.'\\\"/', $config, $commands);
				for($i = 0; $i < sizeof($commands[1]); $i++) {
					echo "  -> ".$commands[1][$i]."\r\n";
				}
			} else { 
				// Display Activities and prepare storage
				echo "\r\nActivity: ".$tags[1][$tagId]." -> Id: ".$tags[2][$tagId]."\r\n";
				if ($displayOnly) echo " debug: tag3: ".$tags[3][$tagId]."\r\n";  // For debug. Should be "activityTypeDisplayName"
				$activities[$tags[2][$tagId]] = $tags[1][$tagId];
				// Alternative 1: Create boolean Activity Variables in a dedicated Category. To enable, comment all lines for Alternative 2. 
				// REMOVED
				
				// Alternative 2: Create one integer Activity Variable with a Profile including different activities. Create only the profile here.
				IPS_SetVariableProfileAssociation($profileNameHarmonyActivity, (int)$tags[2][$tagId], $tags[1][$tagId], "", -1);
			}
			if (!$displayOnly) {
				// Alternative 2: Create the Variable here, and associate profile
				$activitiyVariableId = @IPS_GetVariableIDByName("Activity", $parentID);
				if ($activitiyVariableId === false) {
					$activitiyVariableId = IPS_CreateVariable(1);
					IPS_SetName($activitiyVariableId, "Activity");
					IPS_SetParent($activitiyVariableId, $parentID);
				}
				IPS_SetVariableCustomProfile($activitiyVariableId, $profileNameHarmonyActivity);
			}
			
			$tagId++;
		}
		if (!$displayOnly) {
			// Store the Activities map in an IPS Variable (for fast subsequent access)
			IPS_SetVariableProfileAssociation($profileNameHarmonyActivity, -1, "All Off", "", -1); // Add last activity for all off
			$activitiesVariableId = @IPS_GetVariableIDByName("Activities", $parentID);
			if ($activitiesVariableId === false) {
				$activitiesVariableId = IPS_CreateVariable(3);
				IPS_SetName($activitiesVariableId, "Activities");
				IPS_SetParent($activitiesVariableId, $parentID);
				//IPS_SetHidden($activitiesVariableId, true);
			}
			// Serialize and store array
			SetValue($activitiesVariableId, serialize($activities));
		}
	}
}

/**
* Gets the Activity description from its Id in the Harmony Hub configuration
* A Harmony_getConfig() must have been requested before.
*
* @param $id Activity ID
* @return string Activity description
**/
function Harmony_getActivityFromId($id) {
	if ($id != "-1") {
		$parentID = IPS_GetParent($_IPS['SELF']);
		$activitiesVariableId = @IPS_GetVariableIDByName("Activities", $parentID);
		if ($activitiesVariableId === false) {
			IPS_LogMessage("HARMONY XMPP", "ERROR in Harmony_getActivityFromId(): Activities not defined. Run Harmony_getConfig() first. ");
		}
		$activities = unserialize(GetValue($activitiesVariableId));
	
		//print_r($activities);	
		return $activities["$id"];
	} else {
		return "All Off";
	}
}

/**
* Internal Authentication Processing function
**/
function Harmony_processAuth() {
	global $debug;
	
	// If we have been in a Sesssion Auth, we authenticate as guest to get the identity
	$parentID = IPS_GetParent($_IPS['SELF']);
	$inSessionVarId = @IPS_GetVariableIDByName("In Session", $parentID);
	if ($inSessionVarId === false) {
		$inSessionVarId = IPS_CreateVariable(0);
		IPS_SetName($inSessionVarId, "In Session");
		IPS_SetParent($inSessionVarId, $parentID);
		IPS_SetHidden($inSessionVarId, true);
		SetValue($inSessionVarId, true);
	} 
	$inSession = GetValue($inSessionVarId);

	if ($inSession) {
		//XMPP_Auth('guest@x.com', 'guest');
		XMPP_Auth('guest@connect.logitech.com', 'gatorade.'); // Authenticate as guest
		SetValue($inSessionVarId, false);
	} else {
		$identityVariableId = @IPS_GetVariableIDByName("Identity", $parentID);
		if ($identityVariableId === false) {
		   IPS_LogMessage("HARMONY XMPP", "ERROR in Harmony_processAuth(): Identity Variable does not exist!");
		} else {
			$identity = GetValue($identityVariableId);
			XMPP_Auth($identity.'@connect.logitech.com', $identity); // Authenticate as session
		}
		SetValue($inSessionVarId, true);
	}
}

/**
* Internal XMPP processing function
**/
function Harmony_processIQ($xml) {
	global $debug, $configFilePath;
	
	$content = XMPP_getPayload($xml);
	$parentID = IPS_GetParent($_IPS['SELF']);
	//print_r($content);
	
    preg_match('/<([a-z:]+)\s.*<([a-z:]+)\s/', $xml, $tag);
	if (isset($tag[2])) { // to avoid message of "undefined offset"
		if ($tag[2] == "oa"){
			preg_match('/<!\[CDATA\[\s*(.*)\s*/', $xml, $cdata); // <!\[(CDATA)\[\s*(.*?)\s*>
			$data = $cdata[1];
			str_replace(array("\\", "\""), array("", ""), $data);
			
			if (!strpos($data, "</oa>")) { // If we do not have a closing oa tag within the same message, we need to aggregate. Probably a config msg.
				$configVariableId = @IPS_GetVariableIDByName("Harmony Config", $parentID);
				if ($configVariableId === false) {
					$configVariableId = IPS_CreateVariable(3);
					IPS_SetName($configVariableId, "Harmony Config");
					IPS_SetParent($configVariableId, $parentID);
					IPS_SetHidden($configVariableId, true);
				}
				// Store config in String variable (does not seem to work due to size)
				SetValue($configVariableId , "$data");
				// Write config stream to file
				$configFileHandle = fopen("$configFilePath", "w");
				fwrite($configFileHandle, "$data");
				fclose($configFileHandle);
				if ($debug) IPS_LogMessage("HARMONY XMPP", "RECV: IQ OA Stanza received: Starting CDATA aggregation...\r\n"); // Info/Query Stanza
			} else { // STEP 6 - We got an identity response message
				if ($debug) IPS_LogMessage("HARMONY XMPP", "RECV: IQ Stanza received: identity = ".$content['identity']." - status = ".$content['status']."\r\n"); // Info/Query Stanza
				$identityVariableId = @IPS_GetVariableIDByName("Identity", $parentID);
				if ($identityVariableId === false) {
					$identityVariableId = IPS_CreateVariable(3);
					IPS_SetName($identityVariableId, "Identity");
					IPS_SetParent($identityVariableId, $parentID);
					IPS_SetHidden($identityVariableId, true);
				}
				// Store Identity in String variable
				SetValue($identityVariableId , $content['identity']);
				
				// STEP 7 - If we did the guest authentication (Session Auth == true), we start again now for the session authentication
				$inSessionVarId = @IPS_GetVariableIDByName("In Session", $parentID);
				if ($inSessionVarId === false) {
					IPS_LogMessage("HARMONY XMPP", "ERROR in Harmony_processIQ(): Session Auth Variable not found (after Session token received)");
				} else {
					if (!GetValue($inSessionVarId)) {
						Harmony_init();
						SetValue($inSessionVarId, true); // We are in the Session Auth Process
					}
				}
			}
		} 
		if ($tag[2] == "bind") { // This is not an OA message, we suppose it is a resource binding or identity request reply
			if (!strpos($xml, "<bind")) {
				if ($debug) IPS_LogMessage("HARMONY XMPP", "RECV: Unknown IQ Stanza\r\n"); // Info/Query Stanza
			} else { // STEP 5 - Binding Response (Continuation from Harmony_Read Script)
				preg_match('/<jid>(.*)<\/jid>/', $xml, $jid);
				if ($debug) IPS_LogMessage("HARMONY XMPP", "RECV: IQ Stanza resource binding result - JID: ".$jid[1]."\r\n");
				// Replace 2 lines below by a proper function to get the User Auth Token Value (IPS Tools Library)
				$tokenVariableId = @IPS_GetVariableIDByName("User Auth Token", $parentID);
				if ($tokenVariableId === false) {
					IPS_LogMessage("HARMONY XMPP", "ERROR in Harmony_processIQ(): User Auth Token not defined (after binding reponse).");
				} else {
					if ($debug) IPS_LogMessage("HARMONY XMPP", "SEND: Sending Session Request");
					XMPP_Session(); // Test: Request session
					IPS_Sleep(200);
					$inSessionVarId = @IPS_GetVariableIDByName("In Session", $parentID);
					if ($inSessionVarId === false) {
						IPS_LogMessage("HARMONY XMPP", "ERROR in Harmony_processIQ(): Session Auth Variable not found (before requesting Session token)");
					} else {
						if (!GetValue($inSessionVarId)) { // We request the Session token only if we are authenticated as guest
							if ($debug) IPS_LogMessage("HARMONY XMPP", "SEND: Sending Session Token Request");
							$UserAuthToken = GetValue($tokenVariableId);
							Harmony_sendSessionTokenRequest($UserAuthToken);
							IPS_Sleep(500); // We need to wait to ensure that we receive the identity back from the server
						}
					}
				}
			}
		} 
	} else { // There is no tag, we continue aggregationg the OA data
		$configVariableId = @IPS_GetVariableIDByName("Harmony Config", $parentID);
		if ($configVariableId === false) {
			IPS_LogMessage("HARMONY XMPP", "ERROR in Harmony_processIQ(): Config Variable does not exist!");
			return false;
		} else {
			$isAggregationEnd = false;
			$data = GetValue($configVariableId);
			str_replace(array("\\", "\""), array("", ""), $xml);
			if (!strpos($data, "</oa>")) {
				$data .= $xml;// continue aggregating until we get the closing OA tag
				if (!strpos($xml, "</oa>")) {
					$isAggregationEnd = true;
					if ($debug) IPS_LogMessage("HARMONY XMPP", "RECV: Aggregation of CDATA ended.\r\n");
				} else {
					if ($debug) IPS_LogMessage("HARMONY XMPP", "RECV: Continuing CDATA aggregation...\r\n");
				}
			}
			SetValue($configVariableId , "$data");
			// Write config stream to file
			$configFileHandle = fopen("$configFilePath", "a");
			fwrite($configFileHandle, "$xml");
			fclose($configFileHandle);
			if ($isAggregationEnd) Harmony_formatConfig();// continue aggregating untile we get the closing OA tag
		}
	}
		
}

?>
